//
// "$Id: VncImage.cxx,v 1.2 2004/09/20 06:37:38 dooglio2 Exp $"
//
//	Allows encapsulation of an X display and image.
//

/* Copyright (c) 2003 R. Douglas Barbieri
 *
 * This file is part of VncSelector.
 *
 * VncSelector  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * VncSelector  is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VncSelector ; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Portions also copyright by:
 *
 * (c) Richard J. Moore 1997-2002
 * (c) Matthias Ettrich 2000
 * (c) Aaron J. Seigo 2002
 *
 * (From Ksnapshot: www.kde.org)
 */

/*
Portions Copyright (c) 1999-2001  Ulric Eriksson <ulric@siag.nu>
 
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
 
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
 
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
(From xmagick: http://siag.nu/xmagick/)
*/

#include <stdio.h>
#include <memory.h>

#include "VncImage.h"

VncImage::VncImage() :
        m_image(NULL)
{
    // Empty
}


VncImage::~VncImage()
{
    if( m_image )
    {
        XDestroyImage( m_image );
    }
}



void	VncImage::GetImage( Display* display, XID window )
{
    // Connect to desktop and get window attributes
    //
    if( !window )
    {
        window = DefaultRootWindow( display );
    }

    XWindowAttributes	attributes;
    XGetWindowAttributes( display, window, &attributes );

    const VncRect	rect( 0, 0, attributes.width, attributes.height );

    GC  gc = XCreateGC( display, window, 0, 0 );
    XSetSubwindowMode( display, gc, IncludeInferiors );

#ifdef DEBUG

    printf( "Get XImage from remote desktop\n" );
#endif

    if( m_image )
    {
        XDestroyImage( m_image );
    }

    m_image = XGetImage( display, window,
                         rect.x, rect.y,
                         rect.w, rect.h,
                         AllPlanes,
                         ZPixmap );

    XSetSubwindowMode( display, gc, ClipByChildren );
    XFreeGC( display, gc );
}


void	VncImage::PutImage( Display* display, XID window, const VncRect& rect )
{
    if( !m_image )
    {
        return;
    }

    if( !window )
    {
        window = DefaultRootWindow( display );
    }

    GC  gc = XCreateGC( display, window, 0, 0 );
    XSetSubwindowMode( display, gc, IncludeInferiors );

    XPutImage( display,
               window,
               gc,
               m_image,
               0, 0,
               rect.x, rect.y,
               rect.w, rect.h
             );

    XSetSubwindowMode( display, gc, ClipByChildren );
    XFreeGC( display, gc );
}


void	VncImage::ScaleImage( Display* display, const int new_width, const int new_height )
{
    if( !m_image )
    {
        fprintf( stderr, "No image--call GetImage() or Image( XImage* )\n" );
        return;
    }

    /*
     * Scale algorithm from Ksnapshot (KDE 3)
     */
    const double  r1 = ((double) m_image->height ) / m_image->width;

    int	target_width;
    int target_height;

    if ( r1 * (double) new_width  < (double) new_height )
    {
        target_width 	= new_width;
        target_height = (int) ((double) new_width * r1);
    }
    else
    {
        target_width 	= (int) ((double) new_height / r1);
        target_height = new_height;
    }

    /* Code from Xmagick */
    const int   	format 			= ZPixmap;
    const int   	offset 			= 0;
    char * 			data 			= NULL;
    const int   	bytes_per_line 	= 0;
    const int   	bitmap_pad 		= XBitmapPad     ( display );
    const int		screen 			= DefaultScreen  ( display );
    Visual *		visual 			= DefaultVisual  ( display, screen );
    Colormap 		colormap 		= DefaultColormap( display, screen );
    unsigned int	depth 			= DefaultDepth   ( display, screen );

    XImage* newimage = XCreateImage(
                           display, visual, depth,
                           format, offset, data,
                           target_width, target_height,
                           bitmap_pad, bytes_per_line
                       );

    if( !newimage )
    {
        fprintf( stderr, "Cannot create new image!\n" );
        return;
    }

    const int h_scale = m_image->height / target_height;
    const int w_scale = m_image->width  / target_width;

    newimage->data = (char *) malloc( newimage->bytes_per_line * newimage->height );

#ifdef DEBUG

    printf( "Scale down to new image\n" );
#endif

    for (int y = 0; y < target_height; y++)
    {
        for (int x = 0; x < target_width; x++)
        {
            unsigned long pixel = XGetPixel( m_image, x*w_scale, y*h_scale );
            XPutPixel( newimage, x, y, pixel );
        }
    }

    XDestroyImage( m_image );
    m_image = newimage;
}


void	VncImage::Image( XImage* image )
{
    if( m_image )
    {
        XDestroyImage( m_image );
    }

    m_image = image;
}


XImage*	VncImage::Release()
{
    XImage* retval = m_image;
    m_image = NULL;
    return retval;
}


XImage*	VncImage::Copy()
{
    if( !m_image )
    {
        return NULL;
    }

    XImage*	newimage = (XImage*) malloc( sizeof(XImage) );
    memcpy( newimage, m_image, sizeof(XImage) );

    const int datasize = newimage->bytes_per_line * newimage->height;
    newimage->data = (char *) malloc( datasize );
    memcpy( newimage->data, m_image->data, datasize );

    return newimage;
}


/* $Log: VncImage.cxx,v $
/* Revision 1.2  2004/09/20 06:37:38  dooglio2
/* Adding code to allow thumbnail of logo when no image selected.
/*
 *
 */
