/*******************************************************************************
 * VncSelector.cxx
 *
 * Provides the methods for the VncSelector class.
 *
 * $Id: VncSelector.cxx,v 1.2 2004/09/23 08:37:35 dooglio2 Exp $
 *
 ********************************************************************************/

/* Copyright (c) 2003 R. Douglas Barbieri
 *
 * This file is part of VncSelector.
 *
 * VncSelector  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * VncSelector  is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VncSelector ; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <sys/ptrace.h>
#include <sys/stat.h>
#include <unistd.h>

#include <X11/Xlib.h>

// FLTK
//
#include <FL/x.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Image.H>

#include <dirent.h>
#include <sys/types.h>
#include <sys/wait.h>

// STL
//
#include <vector>

// Local includes
//
#include "VncSelector.h"
#include "VncImage.h"
#include "VncOptions.h"
#include "VncAddDialogUI.h"

VncSelector*  VncSelector::m_instance = NULL;


// Generate one and only one instance of the class
//
VncSelector* VncSelector::Instance()
{
    if( !m_instance )
    {
        m_instance = new VncSelector;
    }
    return m_instance;
}


// Private contructor
//
VncSelector::VncSelector() :
        VncSelectorUI(),
        m_argv(NULL),
        m_env_array(NULL),
        m_wait(true),
        m_thread(500),
        m_triggerThread(false)
{
#ifdef DEBUG
    printf( "VncSelector::VncSelector()\n" );
#endif

    PopulateEnvMap();
    UpdateEnvArray();

    m_vncdir = m_environment["HOME"] + "/.vnc/";

    m_waitDlg = new VncWaitDlg;

    // Register client with thread
    //
    m_thread.AddClient( this );

    // Load up options
    //
    VncOptions  options( m_environment["HOME"] + "/.VncSelector" );
    if( options.LoadOptions() )
    {
        options.GetOption( "viewer"   , m_viewer           );
        options.GetOption( "encodings", m_viewer_encodings );
        options.GetOption( "server"   , m_server           );
    }
    else
    {
        printf("error!\n");
        fflush(stdout);
        throw std::string("Cannot open options file!");
    }
}


// Remove the mutexes
//
VncSelector::~VncSelector()
{
#ifdef DEBUG
    printf( "VncSelector::~VncSelector()\n" );
#endif

    m_triggerThread = false;
    ClearArgArray();

    delete m_waitDlg;
    delete m_env_array;
}


void VncSelector::ShowWaitDlg( const bool show )
{
    if( show )
    {
        m_waitDlg->show();
    }
    else
    {
        m_waitDlg->hide();
    }
    Fl::flush();
}


const bool VncSelector::AddRemoteDisplay( const char* display_name )
{
    VncCritSection  section( &m_mutex );

    if( !display_name )
    {
        return false;
    }

#ifdef DEBUG
    printf( "display_name: %s\n", display_name );
#endif

    std::string name( display_name );

    if( (name == "") || (name.find( ":" ) == std::string::npos) )
    {
        return false;
    }

    std::string filename( m_vncdir + display_name + ".pid" );

    FILE* display_id = fopen( filename.c_str(), "w" );

    if( display_id == NULL )
    {
        return false;
    }

    fprintf( display_id, "-1" );
    fclose( display_id );

    return true;
}


void VncSelector::AddRemoteServer()
{
    VncAddDialogUI* addDlg = new VncAddDialogUI;

    addDlg->show();
}


// Method for showing main window via the FLTK framework
//
void VncSelector::show( int argc, char **argv )
{
#ifdef DEBUG
    printf( "in VncSelector::show()\n" );
#endif

    // Show the window via FLTK
    //
    mainWindow->show( argc, argv );

    // Center it
    //
    VncCommon::CenterWindow( mainWindow );
}



void VncSelector::PopulateEnvMap()
{
    // Stick enviroment vars in map
    //
    for( char ** item = __environ; *item; item++ )
    {
        std::string s_item = *item;
        std::string name;
        std::string value;

        const int equals = s_item.find('=');
        name  = s_item.substr( 0, equals );
        value = s_item.substr( equals+1 );
        m_environment[name] = value;
    }
}


void VncSelector::UpdateEnvArray()
{
    char** environ   = new char*[m_environment.size()+1];
    char** p_environ = environ;

    memset( environ, 0, (m_environment.size()+1) * sizeof(char*) );

    for( EnvMap::iterator iter = m_environment.begin();
            iter != m_environment.end();
            iter++, p_environ++ )
    {
        std::string env_str = iter->first + std::string("=") + iter->second;

        if( (*p_environ == NULL) || (strcmp( *p_environ, env_str.c_str() ) != 0) )
        {
            delete *p_environ;
            *p_environ = strdup( env_str.c_str() );
        }
    }

    m_env_array = environ;
}


void VncSelector::UserRunThread()
{
#ifdef DEBUG
    printf( "in VncSelector::UserRunThread()\n" );
#endif

    VncCritSection  section( &m_mutex );

    if( m_triggerThread )
    {
        DoFork();

        m_triggerThread = false;
    }


    static unsigned int count = 0;

    // Reload the session list
    //
    if( (++count % 2) == 0 )
    {
        LoadSessionList();
    }
}


void VncSelector::ReadVncDir()
{
    m_vncPids.clear();

    // Peruse the users .vnc directory
    //
    DIR * dir = opendir( m_vncdir.c_str() );

    if( dir )
    {
        struct dirent* ent = readdir( dir );

        while( ent )
        {
            std::string name( ent->d_name );

            // Regular file
            //
            if( ent->d_type == DT_REG )
            {
                // Look for .pid extension
                //
                const int pid_pos = name.find( ".pid" );
                if( pid_pos != name.npos  )
                {
                    // If found, strip the .pid extension
                    //
                    VncPid  pidRec;
                    pidRec.m_fileName = m_vncdir + name;
                    pidRec.m_display  = name.substr( 0, pid_pos );

                    FILE* id = fopen( pidRec.m_fileName.c_str(), "r" );

                    if( id )
                    {
                        fscanf( id, "%d", &pidRec.m_pid );
                        fclose( id );

                        m_vncPids.push_back( pidRec );
                    }
                }
            }

            // Next entry
            //
            ent = readdir( dir );
        }

        closedir( dir );
    }
}


// Load all sessions into working GUI list. We are trusting
// that the .pid files living the the user's .vnc directory
// are correct and not stale.
//
// We use the STL string class to excise .pid extensions
// and the POSIX method of perusing directories.
//
// TODO: Add verification to make sure the process is actually
// running.
//
void VncSelector::LoadSessionList()
{
    //VncCritSection  section( &m_mutex );

#ifdef DEBUG
    printf("starting VncSelector::LoadSessionList()\n");
#endif

    int prev_selected = 1;

    // If we have entries in the list, let's preserve
    // the user's selection.
    //
    if( list_sessions->size() )
    {
        prev_selected = list_sessions->value();
        list_sessions->clear();
    }

    // Peruse the users .vnc directory
    //
    ReadVncDir();

    for( VncPidVector::iterator iter = m_vncPids.begin();
            iter != m_vncPids.end();
            iter++ )
    {
        // TODO: verify that the process is indeed still running...
        //
        list_sessions->add
        ( iter->m_display.c_str() );
    }

    // Put the user's selection back
    //
    if( list_sessions->size() )
    {
        // Remember the last selected, unless it got deleted...
        //
#ifdef DEBUG
        printf( "prev_selected: %d\n", prev_selected );
#endif

        prev_selected = (prev_selected <= list_sessions->size())
                        ? prev_selected
                        : list_sessions->size();
        list_sessions->select( prev_selected, 1 );

        // Regenerate the thumbnail of the selected item
        //
        ListClick();
    }
    else
    {
        GenerateThumbnail();
    }
#ifdef DEBUG
    printf("exiting VncSelector::LoadSessionList()\n");
#endif
}


void VncSelector::ListClick()
{
#ifdef DEBUG
    printf( "In ListClick()\n" );
#endif

    const int selected = list_sessions->value();

    if( selected )
    {
        btn_connect->activate();
        btn_kill_session->activate();
    }
    else
    {
        btn_connect->deactivate();
        btn_kill_session->deactivate();
    }

    GenerateThumbnail();
}



// Allows the user to connect to a VNC session
//
void VncSelector::ConnectToSelection()
{
    const int selected = list_sessions->value();

    if( selected )
    {
        const char* display_name = list_sessions->text( selected );

        // Call the VNC viewer app, using the password file
        // which should be living in the user's vnc dir
        //
        const std::string passwd_file = m_vncdir + "/passwd";

        // Create the arguments for the bin to be spawned
        //
        char **argv = new char*[8];
        argv[0] = strdup( m_viewer.c_str() );
        argv[1] = strdup( "-fullscreen" );
        argv[2] = strdup( "-passwd" );
        argv[3] = strdup( passwd_file.c_str() );
        argv[4] = strdup( "-encodings" );
        argv[5] = strdup( m_viewer_encodings.c_str() );
        argv[6] = strdup( display_name );
        argv[7] = NULL;

        // This starts the thread runner
        //
        TriggerThread( argv );
    }
}

// Kill a running instance of the VNC server
//
void VncSelector::KillSelection()
{
#ifdef DEBUG
    printf( "KillSelection()\n" );
#endif

    const int selected = list_sessions->value();

    if( selected )
    {
        // Get the DISPLAY instance
        //
        std::string display_name( list_sessions->text( selected ) );
        display_name = display_name.substr( display_name.find( ':' ) );

        printf( "selected=%d, display_name=%s\n", selected, display_name.c_str() );

        // Kill the instance based on DISPLAY id
        //
        char ** argv = new char*[4];
        argv[0] = strdup( m_server.c_str() );
        argv[1] = strdup( "-kill" );
        argv[2] = strdup( display_name.c_str() );
        argv[3] = NULL;

        // This starts the thread runner
        //
        TriggerThread( argv );
    }
}


// Create a new VNC server instance
//
void VncSelector::CreateSession()
{
#ifdef DEBUG
    printf( "CreateSession()\n" );
#endif

    char ** argv = new char*[2];
    argv[0] = strdup( m_server.c_str() );
    argv[1] = NULL;

    // This starts the thread runner
    //
    TriggerThread( argv );
}


// Clear out the thread array
//
void  VncSelector::ClearArgArray()
{
    if( !m_argv )
    {
        return;
    }

    int argc = 0;
    while( m_argv[argc] )
    {
        delete m_argv[argc++];
    }
    delete m_argv;
    m_argv = NULL;
}


// Trigger the thread
//
void  VncSelector::TriggerThread( char **argv )
{
    VncCritSection  section( &m_mutex );

    // Delete the old argument array
    //
    ClearArgArray();

    // Now assign the new one and trigger the thread
    //
    m_argv = argv;
    m_triggerThread = true;
}


void VncSelector::GenerateThumbnail()
{
    const int selected = list_sessions->value();

    if( !selected )
    {
        group_image->ClearThumbnail();
        return;
    }

    const char* display_name = list_sessions->text( selected );

    Display*    display = XOpenDisplay( display_name );

    if( display )
    {
        VncImage  image;
        image.GetImage( display );
        image.ScaleImage( display, group_image->w(), group_image->h() );
        group_image->ReplaceThumbnail( image.Release() );

        XCloseDisplay( display );
    }
    else
    {
        std::string fileName = m_vncPids[selected-1].m_fileName;
#ifdef DEBUG

        printf( "cannot open remote display for %s\n", display_name );
        printf( "display name: %s\n", XDisplayName( display_name ) );
        printf( "attempting to remove %s\n", fileName.c_str() );
#endif

        unlink( fileName.c_str() );
    }
}



void VncSelector::DoFork()
{
    const int pid = fork();

    // If we are the child process...
    //
    if( pid == 0 )
    {
        struct stat ss;
        const int statretval = stat( m_argv[0], &ss );

        // Santity check--make sure the executable exists and is a regular file!
        //
        if( statretval != 0 || !S_ISREG(ss.st_mode) )
        {
            printf("could not stat() \"%s\"! statretval = %d, errno = %d %s\n",
                   m_argv[0],
                   statretval,
                   errno,
                   strerror(errno) );
        }
        else
        {
            // Good, binary exists, let's run it!
            //
            const int retval = execvp( m_argv[0], m_argv );

            if( retval == -1 )
            {
                printf("m_argv[0] = %s, retval = %d, errno = %d\n", m_argv[0], retval, errno );
                perror("exec failed");
            }
        }

        // Exit on error
        //
        exit(1);
    }

    // Parent process
    //
    // Throw up wait dialog and block app
    //
    ShowWaitDlg();

    if( m_wait )
    {
        // Wait until process has finished
        //
#ifdef DEBUG
        printf("waiting for pid=%d\n", pid );
#endif

        int status = 0;
        waitpid( pid, &status, 0 );

        // TODO: Need some way of informing the user gracefully
        // if the execution of the binary failed...

#ifdef DEBUG

        printf("done waiting; status=%X\n",status);
        printf("\tWIFEXITED=%d\n", WIFEXITED(status));
        printf("\tWEXITSTATUS=%d\n", WEXITSTATUS(status));
        printf("\tWIFSIGNALED=%d\n", WIFSIGNALED(status));
        printf("\tWTERMSIG=%d\n", WTERMSIG(status));
        printf("\tWIFSTOPPED=%d\n", WIFSTOPPED(status));
        printf("\tWSTOPSIG=%d\n", WSTOPSIG(status));
#endif

        m_fork_status = status;
    }

    // Retire please wait dialog and unblock app
    //
    ShowWaitDlg( false );
}


/*
 * $Log: VncSelector.cxx,v $
 * Revision 1.2  2004/09/23 08:37:35  dooglio2
 * Added VncSelector logo so it shows up when there are no sessions loaded.
 *
 * Revision 1.1.1.1  2004/09/19 20:11:40  dooglio2
 * Adding for the first time.
 *
 * Revision 1.1.1.1  2004/09/15 09:43:45  doug
 * Adding version 2 with autoconf support.
 *
 * Revision 1.1.1.1  2004/09/15 09:06:24  doug
 * This is VncSelector 2, with configure support.
 *
 * Revision 1.24  2004/02/07 21:16:03  doug
 * Added config file support.
 *
 * Revision 1.23  2004/01/17 20:08:11  doug
 * Fixed broken mutex code.
 *
 * Revision 1.22  2004/01/17 08:07:17  doug
 * Added button to add a remote server. The user has to "xhost +local_machine"
 * in order for it to work correctly.
 *
 * Revision 1.21  2004/01/13 18:35:49  hajhouse
 * stat() the vncserver binary before execvp()-ing it
 *
 * Revision 1.20  2004/01/13 09:08:58  doug
 * Now handles thumbnailing of desktops fully in background. Now fills
 * in the session list in the background thread as well. Locking is
 * more secure and stable.
 *
 *
 * Revision 1.19  2004/01/11 08:41:02  doug
 * Adding VncThread members to cxx file.
 * VncThread class is no longer a template--can also pass interval
 * (in milliseconds) in constructor.
 *
 * Revision 1.18  2004/01/09 00:04:57  doug
 * Added options for VNCVIEWER (a poor man's configuration).
 *
 * Revision 1.17  2004/01/08 00:10:38  doug
 * Remarked off the VncDisplay class, other changes.
 *
 * Revision 1.16  2004/01/07 00:38:24  doug
 * Moved image manipulation code into a new class, VncImage.
 *
 * Revision 1.15  2004/01/06 19:27:58  doug
 * Thumbnail support is more stable (but still needs a redesign).
 * Retabbed all code to "expandtab."
 * VncThread is now back to being a template (seems to work better).
 *
 * Revision 1.14  2004/01/06 08:16:54  doug
 * Work in progress. Thumbnailing works, but is flakey, particularly with regards
 * to "Kill Session."
 *
 * Revision 1.13  2004/01/04 02:22:39  doug
 * Fixed deadlocking. If sessions exist, list is selected.
 *
 * Revision 1.12  2004/01/03 02:47:37  doug
 * Added thumbnailing support which seems to work correctly.
 *
 * Revision 1.11  2004/01/03 01:55:43  doug
 * Added crude thumbnail support. Work in progress.
 *
 * Revision 1.8  2003/12/27 00:59:24  doug
 * Moved thread code to seperate header file.
 *
 * Revision 1.7  2003/12/18 09:28:30  doug
 * Now correctly supports threads and a "Please Wait" dialog.
 * Centralized the CenterWindow() function into common.h/cxx.
 *
 * Revision 1.6  2003/12/17 20:54:49  doug
 * Work in progress...
 *
 * Revision 1.5  2003/12/16 22:25:06  doug
 * Adding wait dlg at main.
 *
 * Revision 1.4  2003/12/15 20:28:53  doug
 * Work in progress. Added WaitDlg to main class
 *
 * Revision 1.3  2003/12/11 21:15:10  doug
 * Added lots of comments and the GPL header to each file.
 *
 */
