/*******************************************************************************
 * VncThread.cxx
 *
 * Encapsulates the thread logic taken from the main class.
 *
 * $Id: VncThread.cxx,v 1.1.1.1 2004/09/19 20:11:40 dooglio2 Exp $
 *
 ********************************************************************************/

/* Copyright (c) 2003 R. Douglas Barbieri
 *
 * This file is part of VncSelector.
 *
 * VncSelector  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * VncSelector  is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VncSelector ; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "VncThread.h"

///////////////////////////////////////////////////////////////////////////////
// VncMutex
///////////////////////////////////////////////////////////////////////////////

VncMutex::VncMutex()
{
  pthread_mutex_init( &m_mutex, NULL );
}


VncMutex::~VncMutex()
{
  Unlock();
  pthread_mutex_destroy( &m_mutex );
}


void VncMutex::Lock()
{
//#ifdef DEBUG
  //printf( "%p: lock critical section\n", this );
//#endif
    pthread_mutex_lock( &m_mutex );
}

void VncMutex::Unlock()
{
//#ifdef DEBUG
  //printf( "%p: unlock critical section\n", this );
//#endif
    pthread_mutex_unlock( &m_mutex );
}


///////////////////////////////////////////////////////////////////////////////
// VncThread
///////////////////////////////////////////////////////////////////////////////


void* VncThread::thread_runner( void* arg )
{
  VncThread* main = (VncThread*) arg;
  main->RunThread();
  return NULL;
}


// Constructor
//
VncThread::VncThread( const int timeout ) :
  m_timeout( timeout )
{
#ifdef DEBUG
  printf( "VncThread::VncThread(): %p\n", this );
#endif
  pthread_cond_init( &m_cond, NULL );
  pthread_attr_init( &m_attr );
  pthread_create( &m_thread, &m_attr, thread_runner, (void*) this );
}


// Remove the mutexes
//
VncThread::~VncThread()
{
#ifdef DEBUG
  printf( "VncThread::~VncThread(): %p\n", this );
#endif
	m_mutex.Unlock();
  pthread_cond_signal( &m_cond );
  pthread_cond_destroy ( &m_cond );
}


// Thread runner
//
void VncThread::RunThread()
{
#ifdef DEBUG
  printf( "%p: start thread\n", this );
#endif

  struct timespec timeout;

#ifdef DEBUG
  printf( "%p: mutex lock\n", this );
#endif
  m_mutex.Lock();
#ifdef DEBUG
  printf( "%p: after mutex lock\n", this );
#endif

  do
  {
    // Wait .5 seconds
    //
    struct timeval  now;
    gettimeofday( &now, NULL );
    timeout.tv_sec  = now.tv_sec  + 0;
    timeout.tv_nsec = (now.tv_usec + (m_timeout * 1000)) * 1000;

#ifdef DEBUG
    //printf( "." );
    //fflush(stdout);
#endif

    // Iterate through clients and call thread entry point
    //
    ClientList::iterator  iter;

    for( iter = m_clients.begin(); iter != m_clients.end(); iter++ )
    {
      VncThreadClient*  client = *iter;

      if( client )
        client->UserRunThread();
    }
  }
  while( pthread_cond_timedwait( &m_cond, m_mutex.GetMutex(), &timeout ) == ETIMEDOUT );

  m_mutex.Unlock();
#ifdef DEBUG
  printf( "%p: mutex unlock\n", this );
#endif

#ifdef DEBUG
  printf( "%p: thread finish\n", this );
#endif
}


void VncThread::AddClient( VncThreadClient* client )
{
  m_clients.push_back( client );
}


/*
 * $Log: VncThread.cxx,v $
 * Revision 1.1.1.1  2004/09/19 20:11:40  dooglio2
 * Adding for the first time.
 *
 * Revision 1.1.1.1  2004/09/15 09:43:45  doug
 * Adding version 2 with autoconf support.
 *
 * Revision 1.1.1.1  2004/09/15 09:06:24  doug
 * This is VncSelector 2, with configure support.
 *
 * Revision 1.3  2004/01/17 20:08:11  doug
 * Fixed broken mutex code.
 *
 * Revision 1.2  2004/01/13 09:08:58  doug
 * Now handles thumbnailing of desktops fully in background. Now fills
 * in the session list in the background thread as well. Locking is
 * more secure and stable.
 *
 * Revision 1.1  2004/01/11 08:41:02  doug
 * Adding VncThread members to cxx file.
 * VncThread class is no longer a template--can also pass interval
 * (in milliseconds) in constructor.
 *
 */
