/*******************************************************************************
 * VncThread.h
 *
 * Encapsulates the thread logic taken from the main class.
 *
 * $Id: VncThread.h,v 1.1.1.1 2004/09/19 20:11:40 dooglio2 Exp $
 *
 ********************************************************************************/

/* Copyright (c) 2003 R. Douglas Barbieri
 *
 * This file is part of VncSelector.
 *
 * VncSelector  is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * VncSelector  is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VncSelector ; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __VNC_THREAD_H__
#define __VNC_THREAD_H__

#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>
#include <vector>

// class VncMutex
//
// This class handles critical sections
// under-the-hood. When the class is created,
// the critical section is entered. Upon destruction,
// the critical section is then ended.
//
class VncMutex
{
  private:
    pthread_mutex_t m_mutex;

  public:
    VncMutex();
    virtual ~VncMutex();

    void Lock();
    void Unlock();

    pthread_mutex_t*  GetMutex() { return &m_mutex; }
};


class VncCritSection
{
  private:
    VncMutex* m_mutex;

  public:
    VncCritSection( VncMutex* mutex ) :
      m_mutex( mutex )
    {
      if( m_mutex )
			{
        m_mutex->Lock();
			}
    }

    ~VncCritSection()
    {
      if( m_mutex )
			{
        m_mutex->Unlock();
			}
    }
};


class VncThreadClient
{
  public:
    VncThreadClient() {}

    virtual void UserRunThread() = 0;
};


// class VncThread
//
// This class encapulates what it takes to create a thread
// for local use.
//
class VncThread
{
  private:
    typedef std::vector<VncThreadClient*> ClientList;

    ClientList          m_clients;
    pthread_cond_t      m_cond;
    pthread_attr_t      m_attr;
    pthread_t           m_thread;

    int                 m_timeout;
    VncMutex            m_mutex;

    // Private constructors and default assignment
    // operators
    //
    VncThread( const VncThread& );
    void operator =( const VncThread& );

    // Private operations
    /////////////////////

    static void* thread_runner( void* arg );

    // Thread runner
    //
    void RunThread();


  public:
    // Constructor
    //
    VncThread( const int timeout /* in milliseconds */ );

    // Remove the mutexes
    //
    ~VncThread();

    void AddClient( VncThreadClient* client );
};


/*
 * $Log: VncThread.h,v $
 * Revision 1.1.1.1  2004/09/19 20:11:40  dooglio2
 * Adding for the first time.
 *
 * Revision 1.1.1.1  2004/09/15 09:43:45  doug
 * Adding version 2 with autoconf support.
 *
 * Revision 1.1.1.1  2004/09/15 09:06:25  doug
 * This is VncSelector 2, with configure support.
 *
 * Revision 1.8  2004/01/17 20:08:11  doug
 * Fixed broken mutex code.
 *
 * Revision 1.7  2004/01/13 09:08:58  doug
 * Now handles thumbnailing of desktops fully in background. Now fills
 * in the session list in the background thread as well. Locking is
 * more secure and stable.
 *
 * Revision 1.6  2004/01/11 08:41:02  doug
 * Adding VncThread members to cxx file.
 * VncThread class is no longer a template--can also pass interval
 * (in milliseconds) in constructor.
 *
 * Revision 1.5  2004/01/06 19:27:58  doug
 * Thumbnail support is more stable (but still needs a redesign).
 * Retabbed all code to "expandtab."
 * VncThread is now back to being a template (seems to work better).
 *
 * Revision 1.4  2004/01/06 08:16:54  doug
 * Work in progress. Thumbnailing works, but is flakey, particularly with regards
 * to "Kill Session."
 *
 * Revision 1.3  2004/01/04 02:22:39  doug
 * Fixed deadlocking. If sessions exist, list is selected.
 *
 * Revision 1.2  2004/01/03 01:55:43  doug
 * Added crude thumbnail support. Work in progress.
 *
 * Revision 1.1  2003/12/27 00:59:24  doug
 * Moved thread code to seperate header file.
 *
 */
#endif // __VNC_THREAD_H__
